/*****************************************************************************
*                                                                            *
*    Program:     UPSD (UPS Daemon)                                          *
*                                                                            *
*    Date:        Mon Dec 11 14:17:31 WET 1995                               *
*                                                                            *
*    Author:      Alexis Huxley                                              *
*                                                                            *
*    Description: Daemon program to monitor the state of DCD in a            *
*                 specified serial port. This program is compatible with     *
*                 any UPS that has a relay that closes contacts in the event *
*                 of loss of mains power, or a UPS that displays relay-like  *
*                 behaviour.                                                 *
*                                                                            *
*                 It is known to work with the following UPSs with the       *
*                 following cables:                                          *
*                                                                            *
*                 Victron Micro 106                                          *
*                     25 female			              9 female       *
*                     8  <---------------------------------------->  5       *
*                     20 <---------------------------------------->  9       *
*                                                                            *
*****************************************************************************/

/*
 *  Include files and reasons for inclusion
 */

#include <stdio.h>				/* principally for NULL */
#include <errno.h>				/* for errno */
#include <strings.h>				/* for strrchr(), strtok() */
#include <sys/ioctl.h>				/* for ioctl() */
#include <fcntl.h>				/* for O_RDWR in open() */
#include <time.h>				/* for time_t ret'd by time() */
#include <sys/types.h>				/* for stat.h, wait.h */
#include <sys/stat.h>				/* for stat(), umask() */
#include <unistd.h>				/* for sleep(), getpid() */
#include <stdarg.h>				/* for messaging funcs */
#include <sys/wait.h>				/* for wait() */
#include <signal.h>				/* for signal() */
#include <syslog.h>				/* for syslog() */
#if linux
#include <stdlib.h>				/* for setenv() */
#endif
#include <pwd.h>				/* for getpwnam() */

#include "utils.h"				/* for local stuff */

/*
 *  Non-configurable symbols
 */

#define NULLCP ((char *) NULL)			/* I pine for K&R C! */
#define NULLFP ((FILE *) NULL)			/* I pine for K&R C! */
#if linux
#define WALL_CMD  "/usr/bin/wall"		/* Linux wall */
#define WRITE_CMD "/usr/bin/write"		/* Linux write */
#define SHUTDOWN_CMD "/sbin/shutdown"		/* Linux shutdown */
#else
#define WALL_CMD  "/bin/wall"			/* Ultrix wall */
#define WRITE_CMD "/bin/write"			/* Ultrix write */
#define SHUTDOWN_CMD "/etc/shutdown"		/* Ultrix shutdown */
#endif
#define UPSD_ENV_INTERVAL "UPSD_INTERVAL"	/* name of env variable */
#define UPSD_ENV_STATE    "UPSD_STATE"	        /* name of env variable */

/*
 *  Configurable symbols
 */

#define DFLT_DEBUG_LEVEL 2			/* default debug level */
#define DFLT_SHUTDOWN_INTERVAL 240		/* default 4 minute warning */
#define DFLT_WARNING_INTERVAL 30		/* wall every 10 secs */
#define DFLT_CHECKING_INTERVAL 1		/* check every 5 seconds */
#define RESET_FLOW_CONTROL TRUE			/* if badly behaved set this */
#define TEST_FAKE_DCD FALSE 			/* dcd or file /tmp/powerfail */
#define TEST_DONT_WALL FALSE			/* wall or write */
#define TEST_DONT_HALT FALSE 			/* shutdown or pretend */

/*
 *  Forward references for internal functons 
 */

static void usage(void);			/* tell user what to do */
static void shutdown_now(char *); 		/* do an immediate shutdown */
static int  shutdown_warning(char *, int);	/* warn of impending shutdown */
static int  shutdown_cancelled(char *);		/* tell shutdown cancelled */
static int  shutdown_powerfail(void);		/* call as power fails */
static int  callprog(char *, char *, char *, int); /* call external prog/dflt */
static int gen_lockfile_name(char *, char *);   /* write filename into buffer */

/*
 *  Forward references for standard library functions not mentioned in headers
 */

#if !linux					/* linux declares all these */
extern void openlog(char *, int);		/* not mentioned in syslog.h */
extern void syslog(int, char *, ...);		/* not mentioned in syslog.h */
extern int  ioctl(int, int, char *);		/* not mentioned in ioctl.h */
extern void setenv(char *, char *, int);	/* not mentioned in anything! */
#endif

/*
 *  Global variables 
 */

/*
 *  Static Variables
 */

						/* RCS version control id */
static char *rcs_id_upsd_c = "$Header: /diskb/home/alexis/dev/supported/upsd/bin/RCS/upsd.c,v 1.3 1999/01/07 13:32:03 alexis Exp $";
static char *version = "1.20.0.2";	/* version of this file */

/**/

/******************************************************************************
*
*    Function:     main()
*
*    Parameters:   argc - count of command line arguments
*
*    Return value: argv - char vectort for all command line arguments
*
*    Description:  this is the entry point for the UPS daemon. It enters
*                  a loop where is monitors changes in the state of the 
*                  DCD line and, interpreting these as changes in the
*                  state of the mains power supply relayed by the UPS, 
*                  in the event of mains power loss it informs the users
*                  at regular intervals of an impending shutdown and 
*                  finally shuts down the machine.
*                  Of course it handles option processing, locking, signal
*                  processing, logging, debugging, verbose and quiet 
*                  behaviour, etc.
*
******************************************************************************/

int main(
int 	argc,					/* argument count */
char   *argv[])					/* argument list */
{
    char *cp;					/* handy char pointer */
    time_t tm;					/* handy time var */
    char *devcp;				/* pointer to device name */
    int shutdown;				/* flag meaning now shutdown */
    int pwrstat;				/* mains status */
    int lastpwrstat;				/* previous mains status */
#if !TEST_FAKE_DCD
#if linux
    unsigned int ioctl_flags;			/* linux ioctl uses 32 bits */
#else
    unsigned int ioctl_flags;			/* ultrix ioctl uses 32 bits */
#endif
#endif
    int devfd;					/* to open port */
    time_t shutdown_time;			/* time to shutdown at */
    time_t warnmsg_time;			/* time to warn users */
    struct stat devstat;			/* to check dev is serial */
    int shutdown_interval;			/* from powerloss to shutdown */
    unsigned int checking_interval;		/* between power checks */
    int warning_interval;			/* between shutdown warnings */
    char *warning_prog;				/* to run at warning time */
    char *shutdown_prog;			/* to run at shutdown time */
    char *cancelled_prog;			/* to run at cancelled time */
    pid_t lockingpid;				/* active pid holding lock */
#if RESET_FLOW_CONTROL
    char killpwr_bit = TIOCM_RTS;		/* may be superflous */
#endif
    
    progname = mybasename(argv[0], ""); 	/* get basename of this prog */
    shutdown_interval = DFLT_SHUTDOWN_INTERVAL;	/* set default shutdown int */
    checking_interval = DFLT_CHECKING_INTERVAL;	/* set default check interval */
    warning_interval  = DFLT_WARNING_INTERVAL;	/* set default warning int */
    shutdown_prog   = NULLCP;			/* action handled internally */
    warning_prog    = NULLCP;			/* action handled internally */
    cancelled_prog  = NULLCP;			/* action handled internally */

    /**************************************************************************
    *
    *    Process options                        
    *
    **************************************************************************/

    while (argc > 1 && argv[1][0] == '-') {	/* deal with options */
        if (!strcmp(argv[1], "-V")) {		/* -V prints version info */
            printf("%s version %s\n", progname, version);
            return(0);				

        } else if (!strcmp(argv[1], "-v")) {	/* verbose */
            debug_level = 3;

        } else if (!strcmp(argv[1], "-is")) {	/* shutdown interval */
            if (argc == 2) usage();
            if (sscanf(argv[2], "%d", &shutdown_interval) != 1) usage();
            argc--;
            argv++;

        } else if (!strcmp(argv[1], "-iw")) {	/* warning interval */
            if (argc == 2) usage();
            if (sscanf(argv[2], "%d", &warning_interval) != 1) usage();
            argc--;
            argv++;

        } else if (!strcmp(argv[1], "-ic")) {	/* checking interval */
            if (argc == 2) usage();
            if (sscanf(argv[2], "%d", &checking_interval) != 1) usage();
            argc--;
            argv++;

        } else if (!strcmp(argv[1], "-ps")) {	/* program to run at shutdown */
            if (argc == 2) usage();
            shutdown_prog = argv[2];
            argc--;
            argv++;

        } else if (!strcmp(argv[1], "-pw")) {	/* program to run at warning */
            if (argc == 2) usage();
            warning_prog = argv[2];
            argc--;
            argv++;

        } else if (!strcmp(argv[1], "-pc")) {	/* program to run at cancel */
            if (argc == 2) usage();
            cancelled_prog = argv[2];
            argc--;
            argv++;

        } else if (!strcmp(argv[1], "-d")) {	/* debugging (very verbose) */
            if (argc == 2) usage();
            if (sscanf(argv[2], "%d", &debug_level) != 1) usage();
            argc--;
            argv++;

        } else					/* any other option is error */
            usage();

        argc--;					/* ready for next option */
        argv++;
    }

    /**************************************************************************
    *
    *    Process required paramters                        
    *
    **************************************************************************/

    if (argc != 2) usage();			/* required par is port name */
    devcp = argv[1];				/* save it */

    /**************************************************************************
    *
    *    Sanity checks
    *
    **************************************************************************/

    debug(50, "stat()ing %s ...", devcp);	/* does port exist */
    if (stat(devcp, &devstat)) 			/* is it a char device */
        error("couldn't stat %s", devcp);
    if (!(devstat.st_mode & S_IFCHR)) 
	error("%s is not a character device", devcp);

    if (shutdown_prog != NULLCP) {		/* check shutdown prog */
        if (stat(shutdown_prog, &devstat)) 
            error("couldn't stat %s", shutdown_prog);
        if (!(devstat.st_mode & S_IFREG)) 
            error("%s is not a regular file", shutdown_prog);
        if (!(devstat.st_mode & S_IEXEC))  
            error("%s is not an executable program", shutdown_prog);
    }

    if (warning_prog != NULLCP) {		/* check warning prog */
        if (stat(warning_prog, &devstat)) 
            error("couldn't stat %s", warning_prog);
        if (!(devstat.st_mode & S_IFREG)) 
            error("%s is not a regular file", warning_prog);
        if (!(devstat.st_mode & S_IEXEC))  
            error("%s is not an executable program", warning_prog);
    }

    if (cancelled_prog != NULLCP) {		/* check cancelled prog */
        if (stat(cancelled_prog, &devstat)) 
            error("couldn't stat %s", cancelled_prog);
        if (!(devstat.st_mode & S_IFREG)) 
            error("%s is not a regular file", cancelled_prog);
        if (!(devstat.st_mode & S_IEXEC))  
            error("%s is not an executable program", cancelled_prog);
    }

    /**************************************************************************
    *
    *    Initialise interrupt handling
    *
    **************************************************************************/

    signal(SIGTERM, genericsighandler);
    signal(SIGINT,  genericsighandler);


    /**************************************************************************
    *
    *    Lock port and open
    *
    **************************************************************************/

	    					/* lock the device */
    if ((lockingpid=my_lock(devcp, gen_lockfile_name)) > 0)	
        error("%s is in use by process %d", devcp, lockingpid);
    else if (lockingpid < 0)
        error("can't create lockfile: %s", strerror(errno));
    debug(50, "open()ing %s ...", devcp);	/* open for ioctl()ing */
    if ((devfd = open(devcp, O_RDWR | O_NDELAY)) < 0)
        error("can't open %s", devcp);
#if RESET_FLOW_CONTROL 				/* clear flow control */
    ioctl(devfd, TIOCMBIC, (char *) &killpwr_bit); 	
#endif

    /**************************************************************************
    *
    *    Loop monitoring DCD
    *
    **************************************************************************/

    info("version %s started on %s", version, devcp);
    i_am_a_daemon = TRUE;
    for (shutdown=FALSE,lastpwrstat=TRUE;!shutdown;lastpwrstat=pwrstat) {
        sleep(checking_interval);		/* pause between loops */
#if TEST_FAKE_DCD
        pwrstat = !(!stat("/tmp/powerfail", &devstat));
#else
        ioctl(devfd, TIOCMGET, (char *) &ioctl_flags);	/* get DCD and more! */
        pwrstat = !(ioctl_flags & TIOCM_CD); 	/* filter out all except DCD */
#endif
        debug(20, "powerstat: %d, lastpwrstat: %d", pwrstat, lastpwrstat);

        /*
         *  If the power is on and it was on before then don't do anything.
         */

        if (pwrstat == TRUE && pwrstat == lastpwrstat)
            continue;

        /*
         *  If the power is on, but it wasn't before then cancel the 
         *  impending shutdown.
         */

	if (pwrstat == TRUE) {
            shutdown_cancelled(cancelled_prog);
            continue;
        }

        /*
         *  If the power is off but it was on before, then reset shutdown
         *  counters and ...
         */

        if (pwrstat != lastpwrstat) {
            shutdown_time = (tm=time((time_t *) NULL)) + shutdown_interval;
            warnmsg_time = tm;
            shutdown_powerfail();		/* logs only, warning below */
        }
            

        /*
         *  ... if the power is off, regardless of what it was like before, 
         *  and if shutdown time has not been reached yet, then give the
         *  warning message if it's time for the next one and continue with
         *  with the next check.
         */

        if ((tm=time((time_t *) NULL)) < shutdown_time) {
            debug(100, "time is:         %ld", tm);
            debug(100, "warning time is: %ld", warnmsg_time);
            if (tm >= warnmsg_time) {		/* don't warn each time */
                shutdown_warning(warning_prog, (int) (shutdown_time - tm));
						/* schedule next warning */
                warnmsg_time = tm + warning_interval;
                debug(100, "new wrn time is: %ld", warnmsg_time);
            }
            continue;
        }

 	/*
   	 *  If we get to here then the power is off and the time interval 
 	 *  between a power down and the actual shutdown has elapsed. So we
	 *  just set a flag to get out of this loop. (The first line after
         *  the loop will shutdown the machine.)
	 */

 	shutdown = TRUE;			/* to break out of loop */
    }

    shutdown_now(shutdown_prog);		/* "the time has come ..." */
    return(0);					/* not reached */
}

int gen_lockfile_name(
char *lockfile,
char *lockwhat)
{
#if UUCP_STYLE_LOCKS
    sprintf(lockfile, "%s/LCK..%s", LOCK_DIR, mybasename(lockwhat, ""));
#else
    sprintf(lockfile, "%s/%s.pid", LOCK_DIR, mybasename(lockwhat, ""));
#endif
    return(0);
}

/******************************************************************************
*
*    Function:     usage()
*
*    Parameters:   none
*
*    Return value: doesn't return
*
*    Description:  tells the user how to use this command
*
******************************************************************************/

static void usage()
{
    fprintf(stderr, "Usage: %s -V\n", progname);
    fprintf(stderr, "       %s [ -v | -d level ] [ -i[wcs] secs ... ] [ -p[wcs] prog ... ] port\n", progname);
    fprintf(stderr, "\n");
    exit(1);
}

/******************************************************************************
*
*    Function:     shutdown_powerfail()
*
*    Parameters:   none
*
*    Return value: 0
*
*    Description:  as the power fails there is no action to be taken other
*                  then logging the situation. Warning the users takes
*                  place at regular intervals after the power has failed, 
*                  but there is no specific action associated with the 
*                  failure itself.
*
******************************************************************************/

static int shutdown_powerfail()
{
    warning("power has failed");		/* just log it, no action */
    return(0);
}

/******************************************************************************
*
*    Function:     shutdown_warning()
*
*    Parameters:   warning_prog - name of program to execute or NULL
*
*    Return value: 0 for success
*                  non-zero of unsuccessful
*
*    Description:  if the user has specified a command on the command line
*                  to execute when a warning should be issued then this 
*                  is executed at this time, otherwise a message is sent
*                  to all users alerting them that a shutdown will follow
*                  shortly.
*
******************************************************************************/

static int shutdown_warning(
char *warning_prog,
int interval)
{
    warning("shutdown in %d seconds", interval);
    return(callprog(warning_prog, "\n\n***      POWER FAILURE!      ***\n*** SHUTDOWN IN %3d SECONDS! ***\n\n", "impending", interval));
}

/******************************************************************************
*
*    Function:     shutdown_cancelled()
*
*    Parameters:   cancelled_prog - name of program to execute or NULL
*
*    Return value: 0 for success
*                  non-zero of unsuccessful
*
*    Description:  if the user has specified a command on the command line
*                  to execute when a shutdown has been cancelled then this 
*                  is executed at this time, otherwise a message is sent
*                  to all users alerting them that a scheduled shutdown has 
*                  been cancelled.
*
******************************************************************************/

static int shutdown_cancelled(
char *cancelled_prog)
{
    warning("power restored");
    return(callprog(cancelled_prog, "\n\n***      POWER RESTORED!     ***\n***    SHUTDOWN CANCELLED!   ***\n\n", "cancelled", 0));
}

/******************************************************************************
*
*    Function:     shutdown_now()
*
*    Parameters:   shutdown_prog - path of program to execute or NULL
*
*    Return value: doesn't return
*
*    Description:  this program is responsible for actually performing an
*                  immediate shutdown. If prog is specified then it is called,
*                  otherwise "shutdown -h now" is called.
*
******************************************************************************/

static void shutdown_now(
char *shutdown_prog)
{
    warning("shutting down now!");
    if (shutdown_prog == NULL) {		/* if not spec'd, do shutdown */
#if TEST_DONT_HALT
         error("shutdown not enabled, exiting");
#else
         execl(SHUTDOWN_CMD, SHUTDOWN_CMD, "-h", "now", NULLCP);
#endif
    } else {					/* exec user def'd action */
         setenv("UPSD_INTERVAL", "0", 1);
         setenv("UPSD_STATE", "immediate", 1);
         execl(shutdown_prog, shutdown_prog, NULLCP);
    }
}

/******************************************************************************
*
*    Function:     callprog()
*
*    Parameters:   prog - path of program to execute or NULL
*                  fmtstr - message written to all users if prog was NULL
*                  state - word written to environment if prog not NULL
*                  interval - number written to enviroment if prog not NULL
*
*    Return value: 0 for success
*                  exits via error() otherwise
*
*    Description:  If prog is NULL then this function sends the specified 
*                  message to all users. Otherwise it calls prog, first
*                  writing the passed word and number into the enviroment
*                  available to prog.
*
******************************************************************************/

static int callprog(
char *prog,					/* user def'd program or NULL */
char *fmtstr,					/* if no user def'd, say this */
char *state,					/* to user def'd prog in env */
int interval)					/* to user def'd prog in env */
{
    int pipefds[2];				/* to talk to wall */
    char msgbuf[256];				/* handy char buffer */
    int pid;					/* handy pid */
    char numbuf[16];				/* handy chat buffer */

    if (prog == NULL && pipe(pipefds) < 0)	/* no user def'd & can't pipe */
        error("Can't open pipe");
    if ((pid=fork()) < 0)			/* fork to wall */
        error("Can't fork");

    /**************************************************************************
    *
    *    Child
    *
    **************************************************************************/

    /*    
     *  Child process takes input from parent and sends it to users or simply
     *  calls a user defined program if one was supplied on the command line.
     */

    else if (pid == 0) {			/* fork ok, this is child */
        debug(5, "child running ...");
        if (prog == NULL) {			/* no user def'd program */
            debug(5, "child doesn't need to write pipe");
            close(pipefds[1]);			/* no need to write to parent */
            debug(5, "child setting stdin to read from pipe");
            close(0);				/* stdin comes from parent */
            dup(pipefds[0]);
            close(pipefds[0]);
#if TEST_DONT_WALL				/* wall or write, for testing */
            debug(5, "child execing write ...");
            execl(WRITE_CMD, WRITE_CMD, "alexis", NULLCP);
            error("can't exec %s", WRITE_CMD);
#else
            debug(5, "child execing wall ...");
            execl(WALL_CMD, WALL_CMD, NULLCP);
            error("can't exec %s", WALL_CMD);
#endif
        } else {				/* user def'd program to call */
            sprintf(numbuf, "%d", interval);	/* set up env for user def'd */
            setenv(UPSD_ENV_INTERVAL, numbuf, 1);
            setenv(UPSD_ENV_STATE, state, 1);
            execl(prog, prog, NULLCP);		/* exec user defined program */
            error("can't exec %s", prog);
        }
    }

    /**************************************************************************
    *
    *    Parent
    *
    **************************************************************************/

    /*
     *  Meanwhile the parent either feeds a message to the child to be sent
     *  to users or does nothing while the child calls a user defined program.
     *  In either case it waits until the child has finished.
     */

    if (prog == NULL) {				/* no user spec'd prog to run */
        debug(5, "parent doesn't need to read from pipe");
        close(pipefds[0]);			/* no need to read from child */
        debug(5, "parent writing message to pipe ...");
        sprintf(msgbuf, fmtstr, interval);	/* put interval into message */
        write(pipefds[1], msgbuf, (int) strlen(msgbuf)); /* send to child */
        debug(5, "parent closing write side of pipe ...");
        close(pipefds[1]);			/* nothing more to send */
    }
    debug(5, "parent waiting for child to exit");
    while (wait((int *) 0) != pid) 		/* wait for child to exit */
        ;
    debug(5, "parent done");
    return(0);
}
